# coding: UTF-8

'''
訓練・学習
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import tensorflow as tf
import numpy as np

import dataset_loader

# 使用するモデルを指定
from model import easy_gan as model
# from model import dc_gan1 as model
# from model import dc_gan2 as model
# from model import dc_gan3 as model

CAPACITY = 6000
MIN_AFTER_DEQUEUE = 3000

FLAGS = tf.app.flags.FLAGS
tf.app.flags.DEFINE_string('image_base_dir', './data/images',
                           "画像のディレクトリ")
tf.app.flags.DEFINE_string('train_dir', 'train',
                           "学習結果を保存するディレクトリ")

tf.app.flags.DEFINE_float('learning_rate', 0.0001, "学習率")
tf.app.flags.DEFINE_integer('batch_size', 32, "バッチサイズ")
tf.app.flags.DEFINE_integer('max_step', -1, "学習するステップ数")

HOUR_IN_SEC = 60 * 60


def _init_optimizer(learning_rate):
  return tf.train.AdamOptimizer(learning_rate=learning_rate)


def _calc_generator_loss(fake_logits, batch_size):
  cross_entropy = tf.nn.sigmoid_cross_entropy_with_logits(
    labels=tf.ones([batch_size, 1], dtype=tf.float32),
    logits=fake_logits
  )
  cross_entropy_mean = tf.reduce_mean(cross_entropy)
  return cross_entropy_mean


def _calc_discriminator_loss(real_logits, fake_logits, batch_size):
  real_loss = tf.nn.sigmoid_cross_entropy_with_logits(
    labels=tf.ones([batch_size, 1], dtype=tf.float32),
    logits=real_logits
  )
  fake_loss = tf.nn.sigmoid_cross_entropy_with_logits(
    labels=tf.zeros([batch_size, 1], dtype=tf.float32),
    logits=fake_logits
  )
  d_loss = tf.reduce_sum([real_loss, fake_loss])
  return d_loss


def _export_images(fake_batch, batch_size, export_count):
  images = tf.concat(tf.split(fake_batch, batch_size)[:export_count],
                     axis=2)
  images = (tf.squeeze(images, axis=0) + 1.0) / 2.0
  images = tf.image.convert_image_dtype(images, tf.uint8)
  encoded_image = tf.image.encode_png(images)

  return encoded_image


def _write_stat_step(steps, elapsed):
  file_name = '%s-stat.txt' % model.NAME
  file_path = os.path.join(FLAGS.train_dir, file_name)
  message = '%f sec / step, %d steps' % (elapsed, steps)
  print(message)
  with open(file_path, mode='a') as f:
    f.write(message)
    f.write('\n')


def _write_stat_hour(steps, elapsed):
  file_name = '%s-stat.txt' % model.NAME
  file_path = os.path.join(FLAGS.train_dir, file_name)
  message = '%d steps/h, elapsed: %f' % (steps, elapsed)
  with open(file_path, mode='a') as f:
    f.write(message)
    f.write('\n')


def main(args=None):
  assert FLAGS.image_base_dir, \
    'argument --image_base_dir is not set.'
  assert os.path.exists(FLAGS.image_base_dir), \
    '--image_base_dir path %s is not exist.' % FLAGS.image_base_dir

  train_dir = os.path.join(FLAGS.train_dir, model.NAME)
  checkpoint_path = os.path.join(train_dir, 'model.ckpt')

  summary_dir = os.path.join(train_dir, 'train')

  g_step = tf.Variable(initial_value=0, trainable=False)
  d_step = tf.Variable(initial_value=0, trainable=False)

  inputs = tf.random_normal([FLAGS.batch_size, model.INPUT_SIZE])
  fake_batch = model.generator(inputs, FLAGS.batch_size)
  output_images = _export_images(fake_batch, FLAGS.batch_size, 8)

  tf.summary.image('fake_image', fake_batch, max_outputs=4)

  # Generatorが生成する画像のサイズ
  generate_image_size = fake_batch.get_shape()[2]

  image = dataset_loader.load_dataset(
    FLAGS.image_base_dir,
    generate_image_size)

  real_batch = tf.train.shuffle_batch(
    [image],
    FLAGS.batch_size,
    CAPACITY,
    MIN_AFTER_DEQUEUE,
    num_threads=16,
  )

  real_logits = model.discriminator(real_batch, FLAGS.batch_size)
  fake_logits = model.discriminator(fake_batch, FLAGS.batch_size,
                                    is_reuse=True)

  g_loss = _calc_generator_loss(fake_logits, FLAGS.batch_size)
  tf.summary.scalar('g_loss', g_loss)

  d_loss = _calc_discriminator_loss(real_logits, fake_logits,
                                    FLAGS.batch_size)
  tf.summary.scalar('d_loss', d_loss)

  g_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                             scope='generator')
  d_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                             scope='discriminator')

  update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
  with tf.control_dependencies(update_ops):
    g_train_op = _init_optimizer(FLAGS.learning_rate) \
      .minimize(g_loss,
                var_list=g_vars,
                global_step=g_step)
    d_train_op = _init_optimizer(FLAGS.learning_rate) \
      .minimize(d_loss,
                var_list=d_vars,
                global_step=d_step)

  saver = tf.train.Saver(tf.global_variables(),
                         max_to_keep=10)

  summary_op = tf.summary.merge_all()

  coord = tf.train.Coordinator()

  config = tf.ConfigProto(
    allow_soft_placement=True,
    log_device_placement=False)

  with tf.Session(config=config) as sess:
    sess.run(tf.global_variables_initializer())

    summary_writer = tf.summary.FileWriter(summary_dir, sess.graph)

    checkpoint = tf.train.get_checkpoint_state(train_dir)
    if checkpoint and checkpoint.model_checkpoint_path:
      saver.restore(sess, checkpoint.model_checkpoint_path)

    threads = tf.train.start_queue_runners(sess=sess, coord=coord)

    start = time.time()

    try:
      while not coord.should_stop():
        start_step = time.time()

        _, g_loss_value, step = sess.run(
          [g_train_op, g_loss, g_step]
        )

        _, d_loss_value, step = sess.run(
          [d_train_op, d_loss, d_step]
        )

        end_step = time.time()
        elapsed = end_step - start
        elapsed_step = end_step - start_step

        if step % 10000 == 0:
          _write_stat_step(step, elapsed_step)

        if elapsed >= HOUR_IN_SEC:
          _write_stat_hour(step, elapsed)

        assert not np.isnan(g_loss_value), \
          'Model diverged with g_loss = NaN'
        assert not g_loss_value < 0, \
          'Model broken with g_loss = %f' % g_loss_value

        assert not np.isnan(d_loss_value), \
          'Model diverged with d_loss = NaN'
        assert not d_loss_value < 0, \
          'Model broken with d_loss = %f' % d_loss_value

        if step % 100 == 0:
          print('Step: %d, g_loss: %.4f, d_loss: %.4f elapsed: %f'
                % (step, g_loss_value, d_loss_value, elapsed_step))

        if step % 1000 == 0:
          images = sess.run(output_images)
          file_name = 'progress_%d.png' % step
          with open(os.path.join(train_dir, file_name), mode='wb') as f:
            f.write(images)

          summary_str = sess.run(summary_op)
          summary_writer.add_summary(summary_str, step)

        if step % 10000 == 0:
          saver.save(sess, checkpoint_path, global_step=step)

        if FLAGS.max_step > 0 and step >= FLAGS.max_step:
          print('max step reached.')
          break

    except tf.errors.OutOfRangeError:
      print('Done training -- epoch limit reached')
    finally:
      coord.request_stop()

  # Wait for threads to finish.
  coord.join(threads)

  open(os.path.join(train_dir, 'complete'), mode='w').close()


if __name__ == '__main__':
  tf.app.run()
